import Queue
import threading
import sys
import decorator
import inspect
import types
import functools

# This file provides a class that will wrap an object and serialize
# all calls to its methods.  All calls to that object will be queued
# and executed from a single thread, regardless of which thread makes
# the call.

# Based partially on http://stackoverflow.com/questions/2642515/

class SerializerThread(threading.Thread):
    """Thread that retrieves call information from the queue, makes the
    call, and returns the results."""
    def __init__(self, call_queue):
        threading.Thread.__init__(self)
        self.call_queue = call_queue

    def run(self):
        while True:
            result_queue, func, args, kwargs = self.call_queue.get()
            # Terminate if result_queue is None
            if result_queue is None:
                return
            exception = None
            result = None
            try:
                result = func(*args, **kwargs) # wrapped
            except:
                exception = sys.exc_info()
            # Ensure we delete these before returning a result, so
            # we don't unncessarily hold onto a reference while
            # we're waiting for the next call.
            del func, args, kwargs
            result_queue.put((exception, result))
            del exception, result

def _call_in_thread(__func, __queue, *args, **kwargs):
    """Make a call by putting it in the serialization thread's
    queue and waiting for a response"""
    result_queue = Queue.Queue()
    __queue.put((result_queue, __func, args, kwargs))
    ( exc_info, result ) = result_queue.get()
    if exc_info is None:
        return result
    else:
        raise exc_info[0], exc_info[1], exc_info[2]

def serializer_proxy(__cls, *args, **kwargs):
    """Instantiates the given class with the given arguments,
    and returns a SerializerObjectProxy object that proxies all method
    calls to the new object, as well as attribute retrievals.

    The proxied requests, including instantiation, are performed in a
    single thread and serialized between caller threads.
    """
    class SerializerCallProxy(object):
        def __init__(self, call_queue, func):
            self.call_queue = call_queue
            self.result_queue = Queue.Queue()
            self.func = func
        def __call__(self, *args, **kwargs):
            return _call_in_thread(self.func, self.call_queue,
                                   *args, **kwargs)

    class SerializerObjectProxy(object):
        def __init__(self, __cls, *args, **kwargs):
            self.__s_call_queue = Queue.Queue()
            self.__s_thread = SerializerThread(
                self.__s_call_queue)
            self.__s_thread.daemon = True
            self.__s_thread.start()
            self.__s_object = _call_in_thread(__cls, self.__s_call_queue,
                                              *args, **kwargs)
            self._thread_safe = True

        def __getattr__(self, key):
            attr = getattr(self.__s_object, key)
            if not callable(attr):
                getter = SerializerCallProxy(self.__s_call_queue, getattr)
                return getter(self.__s_object, key)
            return SerializerCallProxy(self.__s_call_queue, attr)

        def __del__(self):
            self.__s_call_queue.put((None, None, None, None))
            self.__s_thread.join()

    return SerializerObjectProxy(__cls, *args, **kwargs)
